/* 
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   -
   - The Original Code is this file as it was released on
   - September 25, 2002 (new file, copied from multiviewsTabs.js).
   -
   - The Initial Developer of the Original Code is HJ van Rantwijk.
   - Portions created by HJ van Rantwijk are Copyright (C) 2001-2007
   - HJ van Rantwijk.  All Rights Reserved.
   -
   - Contributor(s):
   -  Michael Vincent van Rantwijk <mv_van_rantwijk@yahoo.com>
   -
*/

var tabIndex; // added for functions: 
              // mzLoadFolderAsGroup(aFolder, aOpenFlag);
              // OpenBookmarkFolder(element, datasource, tabCount, tabOpen);
              // OpenBookmarkFolderFromResource(datasource, resource, tabCount, tabOpen);

const MZ_OPEN_BOOKMARKS_AS_TABS = 0;
const MZ_OPEN_BOOKMARKS_AS_WINDOWS = 1;
const MZ_OPEN_BOOKMARKS_IN_NEW_WINDOW = 2;


function cleanupForTabGroups(aFolder)
{
  var children = aFolder.childNodes;
  var type;
  // iterate over all menu items
  for (var i = 0; i < children.length; i++) {
    if (children[i].getAttribute("folder")) {
      // we found a bookmark folder so get rid of the bookmark items
      for (i = 0; i < children.length; i++) {
        if (children[i].getAttribute("item"))
          aFolder.removeChild(children[i]);
      }
      // hide 'Open This Folder As Group' and the corresponding separator
      if (!children[0].getAttribute("hidden")) {
        children[0].setAttribute("hidden", "true");
        children[1].setAttribute("hidden", "true");
      }
      break;
    }
  }
}

function mzLoadFolderAsGroup(aFolder, aOpenFlag)
{
  if (!mzConfirmOpeningOfBookmarks(aOpenFlag))
    return;

  var url;
  var dupFound = 0;
  var openAsWindows = (aOpenFlag == MZ_OPEN_BOOKMARKS_AS_WINDOWS);
  var openInNewWindow = (aOpenFlag == MZ_OPEN_BOOKMARKS_IN_NEW_WINDOW);
  var children = aFolder.childNodes;
  var newTabs = mzPreferences.readMyPref("multizilla.newtabfor.groupfolders", "int", 0);
  var closeRemaining = mzPreferences.readMyPref("multizilla.bookmark.close-remaining", "bool", false);
  var gBrowser = openInNewWindow ? openWindowAndReturnBrowser() : getBrowser();
  var numberOfTabs = gBrowser.mTabs.length;

  try {
    if (!openAsWindows && !openInNewWindow && newTabs == 2) {
      window.openDialog("chrome://multiviews/content/multiviewsAskSettings.xul", "_blank", "chrome,modal,centerscreen", getBrowser(), closeRemaining);
      newTabs = gBrowser.mNewTabs;
      closeRemaining = gBrowser.mRemoveRemaining;
      if (gBrowser.mNewTabs == -1)
        return;
    }
    newTabs = (newTabs == 1);
    var startTab = newTabs ? numberOfTabs : 0;

    tabIndex = 0;
    /***
      * [0] (Empty)
      * [1] Open this Groupmark/Folder in (new) Tabs
      * [2] Open this Groupmark/Folder in (new) Windows
      * [3] Open this Groupmark/Folder in a new Window
      * [4] Use this Groupmark/Folder as Home Page
      * [5] menuseparator
      */
    for (var i = 6; i < children.length; i++) {
      url = children[i].getAttribute("url");

      if (children[i].getAttribute("container") == "true") {
        OpenBookmarkFolder(children[i], document.getElementById('BookmarksMenu').database, numberOfTabs, newTabs, openAsWindows, openInNewWindow, gBrowser);
      }
      else if (url) {
        if (openAsWindows) {
          window.open(url);
        } 
        else if (!preventDupeTab(url)) {
          if (!newTabs && tabIndex < numberOfTabs) {
            var browser = gBrowser.mTabs[tabIndex].linkedBrowser;
            browser.loadURI(url);
          }
          else {
            gBrowser.addOrGetTab(url);
          }
          tabIndex++;
        }
        else dupFound++;
      }
    }
    // Select the first tab in the group.
    if (!openAsWindows && !openInNewWindow)
      gBrowser.selectedTab = gBrowser.mTabs[startTab];
    if (!newWindows && closeRemaining) {
      if (newTabs) {
        /* startTab = gBrowser.mTabs[startTab].ordinal-1;
        for (i = startTab; i >= 0; i--)
          gBrowser.removeTab(gBrowser.mTabs[i]); */
      }
      else if (dupFound == 0) {
        for (i = numberOfTabs-1; i >= tabIndex; i--)
          gBrowser.removeTab(gBrowser.mTabs[i]);
      }
    }
    gBrowser = getBrowser();
  } catch(ex) {
    gBrowser = getBrowser();
  }
}
////////////////////////////////////////////////////////////////////////////////
// this function opens a bookmark folder, it is a modified copy of
// OpenBookmarkGroup(element, datasource) in navigator.js
////////////////////////////////////////////////////////////////////////////////
function OpenBookmarkFolder(element, datasource, numberOfTabs, newTabs, newWindows, isChild, aTabbrowser)
{
  if (!datasource)
    return;
  var id = element.getAttribute("id");
  var resource = RDF.GetResource(id, true);
  OpenBookmarkFolderFromResource(datasource, resource, numberOfTabs, newTabs, newWindows, isChild, aTabbrowser);
}
////////////////////////////////////////////////////////////////////////////////
// this function opens nested bookmark folders in a new tabs, it is a modified 
// copy of OpenBookmarkGroupFromResource(element, datasource) in navigator.js
////////////////////////////////////////////////////////////////////////////////
function OpenBookmarkFolderFromResource(datasource, resource, numberOfTabs, newTabs, newWindows, isChild, aTabbrowser)
{
  var urlProperty = RDF.GetResource("http://home.netscape.com/NC-rdf#URL");
  var rdfContainer = Components.classes["@mozilla.org/rdf/container;1"].getService(Components.interfaces.nsIRDFContainer);
  rdfContainer.Init(datasource, resource);
  var containerChildren = rdfContainer.GetElements();
  var gBrowser = aTabbrowser ? aTabbrowser : getBrowser();
  tabIndex = 0;

  while (containerChildren.hasMoreElements())
  {
    resource = containerChildren.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
    var target = datasource.GetTarget(resource, urlProperty, true);

    if (target) {
      var url = target.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

      if (url) {
        if (newWindows) {
          window.open(url);
        } 
        else {
          if (!preventDupeTab(url)) {
            var browser = checkForBlankTab();

            if (!isChild && !newTabs && tabIndex < numberOfTabs) {
              browser = gBrowser.mTabs[tabIndex].linkedBrowser;
              browser.loadURI(url);
            }
            else {
              if (browser && mzPreferences.readMyPref("multizilla.tabs.blank-tab-first", "bool", true))
                browser.loadURI(url);
              else
                gBrowser.addTab(url);
            }
            tabIndex++;
          }
        }
      }
    }
    else {
      OpenBookmarkFolderFromResource(datasource, resource, numberOfTabs, newTabs, newWindows, isChild, aTabbrowser);
    }
  }
}

function setStartupGroupmark(bookmarkID)
{
  mzPreferences.writePref("string", "multizilla.tabs.groupmark", bookmarkID);
  mzPreferences.writePref("int", "browser.startup.page", 3);
}

function setStartupBookmark(url)
{
  mzPreferences.writePref("string", "browser.startup.homepage", url);
  mzPreferences.writePref("int", "browser.startup.page", 1);
}

function refreshBookmarkMenuPopup(aParent, aPrefIdentifier)
{
  if (!aParent || !aPrefIdentifier)
    return;

  var menu = aParent;
  var menuItems = menu.childNodes.length;
  var prefValue = Number(mzPreferences.readMyPref(aPrefIdentifier, "int", 0));
  var i = 1;

  if (menuItems > 6) {
    menu.firstChild.setAttribute("hidden", "true");

    for (var b = 1; b < 16; b <<= 1)
      menu.childNodes[i++].setAttribute("hidden", (prefValue & b) ? "false" : "true");
    menu.childNodes[i].setAttribute("hidden", (prefValue & (b-1)) ? "false" : "true");
  }
  if (menuItems == 6 || menuItems == 7 && menu.childNodes[5].localName == "menuitem") {
    for (i = 1; i < 6; i++)
      menu.childNodes[i].setAttribute("hidden", "true");
  }
}

function initializeBookmarksTemplate()
{
  // Bookmark Menu modifications
  var mzPlainFolder = document.getElementById('mzPlainFolder');
  var mzPlainGroup = document.getElementById('mzPlainGroup');
  // Bookmarks Button modifications
  var mzBBPlainFolder = document.getElementById('mzBBPlainFolder');
  var mzBBPlainGroup = document.getElementById('mzBBPlainGroup');

  var prefValue = mzPreferences.readMyPref("multizilla.bookmark.folders", "int", 15);

  if (prefValue && prefValue != 16) {
    mzPlainFolder.setAttribute("disabled", "true");
    mzBBPlainFolder.setAttribute("disabled", "true");
  }
  prefValue = mzPreferences.readMyPref("multizilla.bookmark.groups", "int", 15);

  if (prefValue && prefValue != 16) {
    mzPlainGroup.setAttribute("disabled", "true");
    mzBBPlainGroup.setAttribute("disabled", "true");
  }
}

function initializePersonalToolbarTemplate()
{
  var mzPTPlainFolder = document.getElementById('mzPTPlainFolder');
  var mzPTPlainGroup = document.getElementById('mzPTPlainGroup');
  var mzPTBPlainFolder = document.getElementById('mzPTBPlainFolder');
  var mzPTBPlainGroup = document.getElementById('mzPTBPlainGroup');
  var prefValue = mzPreferences.readMyPref("multizilla.bookmark.folders", "int", 15);

  if (prefValue && !(prefValue & 16)) {
    mzPTPlainFolder.setAttribute("disabled", "true");
    mzPTBPlainFolder.setAttribute("disabled", "true");
  }
  prefValue = mzPreferences.readMyPref("multizilla.bookmark.groups", "int", 15);

  if (prefValue && !(prefValue & 16)) {
    mzPTPlainGroup.setAttribute("disabled", "true");
    mzPTBPlainGroup.setAttribute("disabled", "true");
  }
}

function mzConfirmOpeningOfBookmarks(aOpenFlag, aWindow)
{
  if (aWindow == undefined)
    aWindow = window;

  var confirmationText = "";
  var prefTag = "";

  switch(aOpenFlag)
  {
    case MZ_OPEN_BOOKMARKS_IN_NEW_WINDOW:
         prefTag = "tabgroup-in-window";
         confirmationText = "confirmOpenBookmarksInNewWindow.label";
         break;
    case MZ_OPEN_BOOKMARKS_AS_WINDOWS:
         prefTag = "tabgroup-as-windows";
         confirmationText = "confirmOpenBookmarksAsWindows.label";
         break;
    default: // MZ_OPEN_BOOKMARKS_AS_TABS
         prefTag = "tabgroup";
         confirmationText = "confirmOpenBookmarksInTabs.label";
  }
  var prefIdentifier = "multizilla.bookmark.open-" + prefTag + ".confirmation";
  var shouldConfirm = mzPreferences.readMyPref(prefIdentifier, "bool", false);

  if (shouldConfirm) {
    var aDialogTitle = mzGetTextFromBundle("generalConfirmAction.title");
    var aDialogText = mzGetTextFromBundle(confirmationText);
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService();
    promptService = promptService.QueryInterface(Components.interfaces.nsIPromptService);
    return promptService.confirm(aWindow, aDialogTitle, aDialogText);
  }
  return true;
}
