/***
  * The Developer of this Code is HJ van Rantwijk. Copyright
  * (C) 2004-2007 by HJ van Rantwijk.  All Rights Reserved.
  */

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const STATE_IDLE = 0;
const STATE_BUSY = 1;

const DEFAULT_LIMIT = 5;
const DEFAULT_INTERVAL = 5; // minutes

const INITIAL_DELAY = 1500;
const NEW_WINDOW_DELAY = 1000;

const VERSION_INFO = "mzWebFeeds version 1.2.h";

function debug(aMsg) {
  if (mzWebFeedService.jsConsoleMessages) {
    aMsg = ("mzWebFeed Service: " + aMsg).replace(/\S{80}/g, "$&\n");
    Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService).logStringMessage(aMsg);
  }
}

/* :::::::: The Web Feed Service ::::::::::::::: */

const mzWebFeedService = {
  initialized: 0,
  mTimer: null, // need to hold on to timer ref
  updateQueue: new Array(),
  pendingUpdates: null,
  subscriptionsDS: null,
  RDF: null,
  RDFCUtils: null,
  NS : "http://multizilla.mozdev.org/rdf#",
  prefBranch : null,
  refreshInterval: DEFAULT_INTERVAL,
  refreshLimit: DEFAULT_LIMIT,
  currentState: STATE_IDLE,
  _webFeedsFile: null,
  _webFeedsFileBackup: null,
  _bundle: null,
  _nsIWindowMediator: null,
  _nsIObserverService: null,
  _nsIIOService: null,
  _nsInsIFileOutputStream: null,
  _nsIDOMSerializer: null,
  feedCache: null,
  categoryCache: null,
  refreshStateCache: null,
  failedRefreshCache: null,
  userAgent: "The MultiZilla Feed Reader/Viewer",
  lastUpdateCheck: 0,
  nextUpdateCheck: 0,
  jsConsoleMessages: false,
  storageNeedsRebuild: false,
  // messages: null,

  observe: function(aSubject, aTopic, aData)
  {
    switch (aTopic) {
      case "app-startup":
        // this.messages = new Array();
        // this.messages.push("app-startup");
        this._nsIObserverService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        this._nsIObserverService.addObserver(this, "xpcom-shutdown", false);
        this._nsIObserverService.addObserver(this, "profile-after-change", false);
        break;

      case "profile-after-change":
        // this.messages.push("profile-after-change");
        this._nsIObserverService.removeObserver(this, "profile-after-change", false);
        this._nsIObserverService.addObserver(this, "domwindowopened", false);
        break;

      case "domwindowopened":
        if (aSubject instanceof Ci.nsIDOMChromeWindow) {
          // debug(this.messages);
          if (this.initialized == 0) {
            this._nsIWindowMediator = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
            this._nsIIOService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
            this._nsInsIFileOutputStream = Cc["@mozilla.org/network/file-output-stream;1"].createInstance(Ci.nsIFileOutputStream);
            this._nsIDOMSerializer = Cc["@mozilla.org/xmlextras/xmlserializer;1"].createInstance(Ci.nsIDOMSerializer);
            this.mTimer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
            this.RDF = Cc["@mozilla.org/rdf/rdf-service;1"].getService(Ci.nsIRDFService);
            this.RDFCUtils = Cc["@mozilla.org/rdf/container-utils;1"].getService(Ci.nsIRDFContainerUtils);
            this._bundle = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService)
                                                                 .createBundle("chrome://multiviews/locale/multizilla.properties");
            this.prefBranch = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch("multizilla.feeds.refresh.");
            this.prefBranch.QueryInterface(Ci.nsIPrefBranch2);

            this.lastUpdateCheck = (this._getPref("lastcheck", 0) * 1000);
            dump("\nlastUpdateCheck: "+ this.lastUpdateCheck);
            this.refreshInterval = (this._getPref("interval", DEFAULT_INTERVAL) * 60000);

            if (this.refreshInterval < 60000)
              this.refreshInterval = 60000;
            else if (this.refreshInterval > 3600000)
              this.refreshInterval = 3600000;

            this.jsConsoleMessages = this._getPref("debug_messages");
            this.refreshLimit = this._getPref("limit", DEFAULT_LIMIT);

            if (this.refreshLimit < 1)
              this.refreshLimit = 1;

            this.prefBranch.addObserver("debug_messages", this, false);
            this.prefBranch.addObserver("limit", this, false);
            this.prefBranch.addObserver("interval", this, false);

            this.mTimer.init(this, INITIAL_DELAY, Ci.nsITimer.TYPE_ONE_SHOT);
          }
          else {
            var now = new Date().getTime();
            var nextCheck = new Date(this.nextUpdateCheck).getTime();

            if ((now + NEW_WINDOW_DELAY + 500) < nextCheck) {
              this.initialized = 2;
              this.mTimer.init(this, NEW_WINDOW_DELAY, Ci.nsITimer.TYPE_ONE_SHOT);
            }
          }
        }
        break;

      case "timer-callback":
        /* if (this.messages) {
          debug(this.messages);
          this.messages = new Array();
        } */
        switch(this.initialized) {
          case 0: 
            if (this.getMostRecentBrowserWindow())
              this.init();
            break;
          case 1: 
            this.checkFeedsForUpdates();
            break;
          case 2:
            this.initialized = 1;
            var currentState = this.getGlobalRefreshState(false);
            this._nsIObserverService.notifyObservers(null, "WebFeedService", currentState);
            break;
        }
        break;

      case "nsPref:changed":
        switch(aData) {
          case "debug_messages":
            this.jsConsoleMessages = this._getPref(aData, this.jsConsoleMessages);
            debug("JS Console messages: " + (this.jsConsoleMessages ? "enabled" : "disabled"));
            break;
          case "limit":
            this.refreshLimit = this._getPref(aData, this.refreshLimit);

            if (this.refreshLimit < 1)
              this.refreshLimit = 1;

            debug("Refresh limit set to: " + this.refreshLimit);
            break;
          case "interval":
            this.refreshInterval = (this._getPref(aData, this.refreshInterval) * 60000);

            if (this.refreshInterval < 60000)
              this.refreshInterval = 60000;
            else if (this.refreshInterval > 3600000)
              this.refreshInterval = 3600000;

            this.mTimer.init(this, this.refreshInterval, Ci.nsITimer.TYPE_ONE_SHOT);
            debug("Refresh interval set to: " + this.refreshInterval + "ms delay");
            break;
          }
        break;

      case "xpcom-shutdown":
        /* 
         * We need to drop our timer reference here to avoid a leak
         * since the timer keeps a reference to the observer! 
         */
        this.mTimer = null;
        this._nsIObserverService.removeObserver(this, "domwindowopened", false);
        this._nsIObserverService.removeObserver(this, "xpcom-shutdown", false);
        break;
    }
  },

  init: function()
  {
    this.initialized = 1;
    this.failedRefreshCache = new Array();
    this.getFeedCategories(); // initializes the 'categoryCache'
    this.getFeedSubscriptions(); // initializes the 'feedCache'
    var currentState = this.getGlobalRefreshState(true);
    this._nsIObserverService.notifyObservers(null, "WebFeedService", currentState);

    var refreshDelay = 2500;
    var now = new Date().getTime();

    if (this.lastUpdateCheck > 0) {
      refreshDelay = (now - this.lastUpdateCheck);

      if (refreshDelay > this.refreshInterval)
        refreshDelay = 2500;
      else if(refreshDelay < this.refreshInterval)
        refreshDelay = (this.refreshInterval - refreshDelay);
    }
    // dump("\nrefreshDelay        : " + refreshDelay);
    this.mTimer.init(this, refreshDelay, Ci.nsITimer.TYPE_ONE_SHOT);
    debug("Initialized with a " + this.refreshInterval + "ms delay");
    // XXX: Check me!!!
    if (this.feedCache.length) {
      this.nextUpdateCheck = new Date(now + refreshDelay);
      debug("Next refresh: " + this.nextUpdateCheck);
    }
  },

  initDataSource: function()
  {
    if (!this.subscriptionsDS)
      this.subscriptionsDS = this.getMostRecentBrowserWindow().mzGetNewOrCurrentDSByFilename("mzWebFeeds.rdf", false, null, false);

    return this.subscriptionsDS;
  },

  checkFeedsForUpdates: function()
  {
    debug("checkFeedsForUpdates()");

    if (this.feedCache.length) {
      debug("Checking idle state...");

      if (this.currentState == STATE_IDLE) {
        debug("Current state: STATE_IDLE");
        this.updateAllFeeds(false);
      }
    }
    else
      this.mTimer.init(this, this.refreshInterval, Ci.nsITimer.TYPE_ONE_SHOT);
  },

  updateFeeds: function(aFeedResourcesString, aForcedCheckFlag)
  {
    if (aFeedResourcesString) {
      debug("Building feed list...")
      var feedResourceList = aFeedResourcesString.split(',');
      this.refreshChannels(feedResourceList, aForcedCheckFlag);
    }
    else
      this.mTimer.init(this, this.refreshInterval, Ci.nsITimer.TYPE_ONE_SHOT);
  },

  updateAllFeeds: function(aForcedCheckFlag)
  {
    var feedSubscriptions = this.getFeedSubscriptions();

    if (feedSubscriptions) {
      debug("Building feed list...")
      this.lastUpdateCheck = new Date().getTime();
      debug("Updating last refresh time...");
      this.prefBranch.setIntPref("lastcheck", (this.lastUpdateCheck / 1000));
      this.refreshChannels(feedSubscriptions, true);
    }
    else
      this.mTimer.init(this, this.refreshInterval, Ci.nsITimer.TYPE_ONE_SHOT);
  },

  refreshChannels: function(aFeedList, aForcedCheckFlag)
  {
    this.currentState == STATE_BUSY;
    this.pendingUpdates = new Array();

    try {

      if (aFeedList.length > 0) {
        var dataSource = this.initDataSource();
        debug("Checking local feed storage...");
        for (var i = 0; i < aFeedList.length; i++) {
          var urn = (typeof aFeedList[i] == "object") ? aFeedList[i].urn : aFeedList[i];
          // dump("\n\ncurrentResource[ " + i + " ]: " + urn);
          var currentResource = this.RDF.GetResource(urn);
          var feedContainer = this.RDFCUtils.MakeSeq(dataSource, currentResource);

          if (feedContainer) {
            var dataObj = new Object();
            dataObj.urn = urn;

            var resource = dataSource.GetTarget(currentResource, this.RDF.GetResource(this.NS + "state"), true);
            // dump("\nChecking 'state': " + resource);
            if (resource)
              dataObj.state = resource.QueryInterface(Ci.nsIRDFLiteral).Value;
            // else
              // dump("\nNo 'state' for: " + urn);

            resource = dataSource.GetTarget(currentResource, this.RDF.GetResource(this.NS + "url"), true);
            // dump("\nChecking 'url': " + resource);
            if (resource)
              dataObj.url = resource.QueryInterface(Ci.nsIRDFLiteral).Value;
            // else
              // dump("\nNo 'url' for: " + urn);

            resource = dataSource.GetTarget(currentResource, this.RDF.GetResource(this.NS + "date"), true);
            // dump("\nChecking 'date': " + resource);
            if (resource)
              dataObj.date = resource.QueryInterface(Ci.nsIRDFLiteral).Value;
            // else
              // dump("\nNo 'date' for: " + urn);

            resource = dataSource.GetTarget(currentResource, this.RDF.GetResource(this.NS + "lastModified"), true);
            // dump("\nChecking 'lastmodified': " + resource);
            if (resource instanceof Ci.nsIRDFNode)
              dataObj.rdfDate = resource.QueryInterface(Ci.nsIRDFLiteral).Value;
            else
              dataObj.rdfDate = dataObj.date;

            dataObj.forcedRefresh = (aForcedCheckFlag);
            dataObj.retries = -1;
            this.pendingUpdates.push(dataObj);
          }
          else {
            debug("Refresh failed for: " + urn);
            var category = this.getCategoryByFeed(this.RDF.GetResource(urn));
            var categoryContainer = this.RDFCUtils.MakeSeq(dataSource, this.RDF.GetResource(category));

            if (categoryContainer) {
              categoryContainer.RemoveElement(currentResource, true);
              this.flushDataSource(dataSource);
            }
            aFeedList = aFeedList.splice(i ,1);
          }
        }
      }
      if (this.pendingUpdates.length) {
        // this.updateQueue = new Array();
        debug(aFeedList.length + " feed(s) pending for a refresh...");
        this.checkPendingUpdates();
      }
      else {
        this.currentState = STATE_IDLE;
        // The update check has finished, forced a re-init of the global refresh state
        var currentState = this.getGlobalRefreshState(true);
        this._nsIObserverService.notifyObservers(null, "WebFeedService", currentState);
        // dump("\nrefreshInterval: " + this.refreshInterval);
        this.mTimer.init(this, this.refreshInterval, Ci.nsITimer.TYPE_ONE_SHOT);
        debug("No new feed items available");
        this.nextUpdateCheck = new Date(this.lastUpdateCheck + this.refreshInterval);
        debug("Next refresh: " + nextUpdateCheck);
      }
    } catch(ex) { debug(ex); }
  },

  checkPendingUpdates: function()
  {
    if (this.pendingUpdates && this.pendingUpdates.length && 
        (this.updateQueue.length <= this.refreshLimit)) {
      // dump("\ncheckPendingUpdate() " + this.pendingUpdates.length + " item(s) left...");
      var _this = this; // for event handlers
      var dataObj = this.pendingUpdates[0];

      if (dataObj && dataObj.urn && dataObj.url) {
        if (dataObj.retries < 3)
          dataObj.retries++;
        else {
          if (this.pendingUpdates.length == 1)
            this.removeFromPendingQueue(dataObj.urn);
          else {
            dataObj = this.pendingUpdates[0];
            this.pendingUpdates.splice(i, 1);
            this.pendingUpdates.push(dataObj);
            dataObj = this.pendingUpdates[0];
          }
          this.checkPendingUpdates();
        }
        // debug("Checking feed: " + dataObj.urn);
        this.addOrChangeProperty(this.subscriptionsDS, this.RDF.GetResource(dataObj.urn), this.RDF.GetResource(this.NS + "state"), "checking");
        var xmlRequest = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
        // xmlRequest.onprogress = function(event) { onXMLProgress(event); };
        xmlRequest.open("HEAD", dataObj.url, true);
        xmlRequest.setRequestHeader("User-Agent", this.userAgent);
        xmlRequest.setRequestHeader("Cache-Control", "no-cache");
        // var lastModified = "Sun, 15 Apr 2007 01:15:00 GMT";
        // xmlRequest.setRequestHeader("If-Modified-Since", lastModified, false);

        xmlRequest.onerror = function(aEvent)
        {
          var request, status;
          var resource = _this.RDF.GetResource(dataObj.urn);
          _this.addOrChangeProperty(_this.subscriptionsDS, resource, _this.RDF.GetResource(_this.NS + "state"), "");
          _this.hasNewArticles(resource, true, null);
          _this.removeFromPendingQueue(dataObj.urn);

          try {
            request = aEvent.target;
            // the following may throw (e.g. a local file or timeout)
            status = request.status;
          }
          catch(ex) {
            request = aEvent.target.channel.QueryInterface(Ci.nsIRequest);
            status = request.status;
          }
          xmlRequest.abort();
        }

        xmlRequest.onload = function()
        {
          // dump("\nxmlRequest.readyState: " + xmlRequest.readyState);
          // dump("\nxmlRequest.status: " + xmlRequest.status);
          if (xmlRequest.readyState == 4) {
            // dump("\n\nxmlRequest.status: " + xmlRequest.status);
            // dump("\ngetAllResponseHeaders: " + xmlRequest.getAllResponseHeaders());
            if (xmlRequest.status == 200) {
              if (dataObj.forcedRefresh) {
                dump("\nForced refresh of: " + dataObj.urn);
                _this.updateStageTwo(dataObj);
              }
              else {
                var lastModifiedDateFromRDF = Number(dataObj.rdfDate);
                // dump("\nlastModifiedDateFromRDF: " + lastModifiedDateFromRDF);
                var invalidRDFDate = isNaN(lastModifiedDateFromRDF);
                // dump("\ninvalidRDFDate: " + invalidRDFDate);
                var lastModifiedDataHeader = xmlRequest.getResponseHeader("Last-Modified");
                var lastModifiedDataFromHeader = lastModifiedDataHeader;
                // dump("\nlastModifiedDataFromHeader-0: " + lastModifiedDataFromHeader);
                var invalidHeaderDate = false;

                if (lastModifiedDataHeader == null)
                  invalidHeaderDate = true;
                else {
                  lastModifiedDataFromHeader = new Date(lastModifiedDataFromHeader).toUTCString();
                  // dump("\nlastModifiedDataFromHeader-: " + lastModifiedDataFromHeader);
                  lastModifiedDataFromHeader = new Date(lastModifiedDataFromHeader).getTime();
                  // dump("\nlastModifiedDataFromHeader-2: " + lastModifiedDataFromHeader);
                  lastModifiedDataFromHeader = Number(lastModifiedDataFromHeader);
                  // dump("\nlastModifiedDataFromHeader-3: " + lastModifiedDataFromHeader);
                  invalidHeaderDate = isNaN(lastModifiedDataFromHeader);

                  // if (invalidHeaderDate)
                    // dump("\nLast-Modified: invalid for: " + dataObj.url);
                }
                // dump("\ninvalidHeaderDate: " + invalidHeaderDate);
                var validDates = (invalidRDFDate == false && invalidHeaderDate == false);
                // dump("\nvalidDates: " + validDates);

                if (validDates && (lastModifiedDateFromRDF > lastModifiedDataFromHeader)) {
                  validDates = false;
                  invalidHeaderDate = true;
                  lastModifiedDataFromHeader = 0;
                  /* dump("\nLast-Modified: ahead of time: " + dataObj.url +
                       "\nLast-Modified: " + (lastModifiedDateFromRDF - lastModifiedDataFromHeader) +
                       "\nlastModifiedDateFromRDF: " + lastModifiedDateFromRDF +
                       "\nlastModifiedDataFromHeader-3: " + lastModifiedDataFromHeader +
                       "\n"); */
                }
                // dump("\nvalidDates: " + validDates);
                dataObj.lastModifiedDate = (invalidHeaderDate) ? 0 : lastModifiedDataFromHeader;
                /* dump("\nChannel urn                   : " + dataObj.urn +
                     "\nHTTP Last-Modified            : " + lastModifiedDataFromHeader +
                     "\nDate stored in mzWebFeeds.rdf : " + lastModifiedDateFromRDF); */

                if ((validDates && (lastModifiedDataFromHeader > lastModifiedDateFromRDF)) || invalidRDFDate || invalidHeaderDate) {
                  _this.updateStageTwo(dataObj);
                }
                else {
                  if (_this.isCheckingState(dataObj.urn))
                    _this.hasNewArticles(dataObj.urn, true, null);
                  else 
                    _this.addOrChangeProperty(this.subscriptionsDS, _this.RDF.GetResource(dataObj.urn), _this.RDF.GetResource(_this.NS + "state"), dataObj.state);
                  _this.removeFromPendingQueue(dataObj.urn);
                }
              }
            }
            else {
              dump("\nERROR: " + xmlRequest.status);
              // _this.addOrChangeProperty(_this.subscriptionsDS, _this.RDF.GetResource(dataObj.urn), _this.RDF.GetResource(_this.NS + "state"), "");
              // _this.hasNewArticles(dataObj.urn, true, null);
              // _this.removeFromPendingQueue(dataObj.urn);
            }
          }
        }
        xmlRequest.send(null);
        this.mTimer.initWithCallback(this, 30000, Ci.nsITimer.TYPE_ONE_SHOT); // call this.notify()
      } // if (dataObj && dataObj.url && dataObj.urn) {
      else {
        dump("\ndataObj.urn: " + dataObj.urn +
             "\ndataObj.url: " + dataObj.url);
        this.removeFromPendingQueue(dataObj.urn);
      }
    } // if (this.pendingUpdates && this.pendingUpdates.length && ...
    else {
      var failedFeeds = this.failedRefreshCache.length;

      if (failedFeeds) {
        debug(failedFeeds + " feed(s) failed. Retrying...");
        var failedFeeds = this.failedRefreshCache;
        this.failedRefreshCache = new Array();
        this.refreshChannels(failedFeeds, true);
      }
      else {
        this.currentState = STATE_IDLE;
        // The update check has finished, forced a re-init of the global refresh state
        var currentState = this.getGlobalRefreshState(true);
        this._nsIObserverService.notifyObservers(null, "WebFeedService", currentState);
        this.mTimer.init(this, this.refreshInterval, Ci.nsITimer.TYPE_ONE_SHOT);
        debug("Feed refresh (check) finished");
        this.nextUpdateCheck = new Date(this.lastUpdateCheck + this.refreshInterval)
        debug("Next refresh: " + this.nextUpdateCheck);
      }
    }
  },

  notify: function(aTimers)
  {
    this.checkPendingUpdates();
  },

  updateStageTwo: function(aDataObj)
  {
    var _this = this;
    var feedURN = aDataObj.urn;
    var xmlRequest = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
    xmlRequest.overrideMimeType("application/xml"); // "application/rss+xml"

    xmlRequest.onprogress = function(aEvent)
    {
      var percentage = (aEvent.position * 100) / aEvent.totalSize;
      // dump("\npercentage: " + percentage);
      // statusMeter.value = percentage;
    }
    xmlRequest.open("GET", aDataObj.url);
    xmlRequest.setRequestHeader("User-Agent", this.userAgent);
    xmlRequest.setRequestHeader("Cache-Control", "no-cache");
    xmlRequest.overrideMimeType("application/xml");

    xmlRequest.onerror = function(aEvent)
    {
      dump("\nupdateStageTwo - onerror");
      var request, status;

      try {
        request = aEvent.target;
        // the following may throw (e.g. a local file or timeout)
        status = request.status;
      }
      catch(ex) {
        request = aEvent.target.channel.QueryInterface(Ci.nsIRequest);
        status = request.status;
      }
      xmlRequest.abort();

      if (_this.failedRefreshCache.indexOf(feedURN) == -1)
        _this.failedRefreshCache.push(feedURN);

      _this.removeFromPendingQueue(feedURN);
    }

    xmlRequest.onload = function()
    {
      var xmldoc = xmlRequest.responseXML;

      if (xmldoc) {
        var dataCollection = _this.getFeedInfoFromDocument(xmldoc);
        // aDataObj.lastModifiedDate = 0 when there was no (valid) date send
        // aDataObj.rdfDate = 0 when we don't have a date (yet) in mzWebFeeds.rdf
        if (aDataObj.forcedRefresh || (aDataObj.lastModifiedDate == 0 || aDataObj.rdfDate == 0) || (dataCollection[7] > aDataObj.rdfDate) ||
           (aDataObj.lastModifiedDate > dataCollection[7]) || (aDataObj.lastModifiedDate > aDataObj.rdfDate))
        {
          dataCollection[1] = aDataObj.url;
          dataCollection[8] = aDataObj.urn;
          dataCollection = _this.getDataCollectionFromDocument(xmldoc, dataCollection);
          // dump("\nUpdate available for: " + aDataObj.urn);
          /***
            * We have a new update available and callback function updateStageThree
            * will remove all current items, and add this update to the update queue
            * which will later be used by callback function finalizeUpdateProcess.
            */
            _this.storeFeed({urn:aDataObj.urn, document:xmldoc});
            _this.updateStageThree(dataCollection);
        }
        else {
          // dump("\naDataObj.state: " + aDataObj.state);
          var newState = (aDataObj.state == "checking") ? "" : aDataObj.state;  
          _this.addOrChangeProperty(_this.subscriptionsDS, _this.RDF.GetResource(aDataObj.urn), _this.RDF.GetResource(_this.NS + "state"), newState);
          _this.removeFromPendingQueue(aDataObj.urn);
        }
      }
    }
    xmlRequest.send(null);
  },

  updateStageThree: function(aDataCollection)
  {
    var urn = aDataCollection[0][8];
    // debug("Refreshing feed: " + urn);
    var resource = this.RDF.GetResource(urn);
    var category = this.getCategoryByFeed(resource);
    var dataSource = this.initDataSource();

    this.removeResources(dataSource, resource); // remove article container
    this.removeDataResources(dataSource, resource, false); // remove articles
    this.addFeedFromDataCollection(aDataCollection, true, category);
    this.removeFromPendingQueue(urn);
  },

  isCheckingState: function(aResourceString)
  {
    var dataSource = this.initDataSource();

    return (dataSource.HasAssertion(this.RDF.GetResource(aResourceString), 
                                    this.RDF.GetResource(this.NS + "state"), 
                                    this.RDF.GetLiteral("checking"), 
                                    true));
  },

  removeFromPendingQueue: function(aTargetURN, aSkipCaller)
  {
    for (i in this.pendingUpdates) {
      var target = (this.pendingUpdates[i].urn == undefined) ? this.pendingUpdates[i][8] : this.pendingUpdates[i].urn;

      if (target == aTargetURN) {
        // dump("\nRemoving: " + aTargetURN + " from pending queue...");
        this.pendingUpdates.splice(i, 1);
      }
    }
    if (this.pendingUpdates.length == 0)
      this.currentState = STATE_IDLE;

    this.checkPendingUpdates();
  },

  hasNewArticles: function(aResource, aUpdateChannelFlag, aTargetState)
  {
    /* dump("\n\nhasNewArticles - aResource         : " + aResource.Value +
              "\nhasNewArticles - aUpdateChannelFlag: " + aUpdateChannelFlag); */
    var dataSource = this.initDataSource();
    var state = (aTargetState == null) ? false : aTargetState;
    var resource = aResource.Value.replace(/:article-\d*/, '');
    resource = this.RDF.GetResource(resource);

    if (aTargetState == null) {
      var channels = this.getResourcesFromContainer(resource);

      if (channels && channels.length) {
        var property = this.RDF.GetResource(this.NS + "state");
        // var targetVisited = this.RDF.GetLiteral("visited");
        // var targetUnread = this.RDF.GetLiteral("unread");

        for (i in channels) {
          // dump("\nchecking: " + channels[i]);
          var feedResource = this.RDF.GetResource(channels[i]);
           /* if (dataSource.HasAssertion(feedResource, property, targetUnread, true) || 
              dataSource.HasAssertion(feedResource, property, targetVisited, false)) {
            state = true;
            break;
          } */
          var data = this.subscriptionsDS.GetTarget(feedResource, property, true);

          if (data) {
            var currentState = data.QueryInterface(Ci.nsIRDFLiteral).Value;
            // dump("\ncurrentState: " + currentState);

            if (currentState == "unread") {
              // dump("\nstate is 'unread'");
              state = true;
              break;
            }
          }
          else {
            // dump("\nno state node: " + data);
            state = true;
            break;
          }
        }
      }
      state = (state) ? "new" : "";
    }
    /* dump("\nhasNewArticles - state             : " + state +
            "\nhasNewArticles - resource          : " + resource.Value); */
    if (aUpdateChannelFlag)
      this.setVisitedState(resource, state);

    // The check has finished, forced a re-init of the global refresh state
    var currentState = this.getGlobalRefreshState(true);
    this._nsIObserverService.notifyObservers(null, "WebFeedService", currentState);
    this.flushDataSource(dataSource);
    return state;
  },

  addFeedFromDataCollection: function(aDataCollection, aRefreshFlag, aCategory)
  {
    this.getMostRecentBrowserWindow().mzFeedHandler.addFeedFromDataCollection(aDataCollection, aRefreshFlag, aCategory);
  },

  addFeedCategory: function(aCategoryName, aTargetResource)
  {
    try {
      var dataSource = this.initDataSource();
      var rootResource = this.RDF.GetResource("urn:feeds:root");
      var rootContainer = this.RDFCUtils.MakeSeq(dataSource, rootResource);

      /* if (!rootContainer) {
        var newContainer = Cc["@mozilla.org/rdf/container;1"].createInstance(Ci.nsIRDFContainer);
        rootContainer = newContainer.Init(dataSource, rootResource);
        // this.flushDataSource(dataSource);
      }
      else { */
        var propertiesArray = [ "name", "group", "date", "state" ];
        var category = (aCategoryName) ? aCategoryName : "New Feed Group";
        var valuesArray = [ category, "true", new Date().getTime(), "undetermined" ];
        var anonymousResource = this.RDF.GetAnonymousResource();
        var resource = this.RDF.GetResource("urn:feedgroup:" + anonymousResource.Value.replace(/^rdf:/, ''));

        for (i in propertiesArray) {
          var property = this.RDF.GetResource(this.NS + propertiesArray[i]);
          dataSource.Assert(resource, property, this.RDF.GetLiteral(valuesArray[i]), true);
        }
        var targetIndex = (aTargetResource == undefined) ? 0 : rootContainer.IndexOf(this.RDF.GetResource(aTargetResource));

        if (targetIndex > 0 && targetIndex < rootContainer.GetCount())
          rootContainer.InsertElementAt(resource, targetIndex, true);
        else
          rootContainer.AppendElement(resource);

        this.RDFCUtils.MakeSeq(dataSource, resource);
        this.flushDataSource(dataSource);
        this.categoryCache.push(resource.Value)
        return resource.Value;
      // }
    }
    catch(ex) { 
      dump("\naddFeedCategory - ex: " + ex);
    }
    return null;
  },

  setVisitedState: function(aResource, aState)
  {
    var dataSource = this.initDataSource();

    if (aState != "")
      this.addOrChangeProperty(dataSource, aResource, this.RDF.GetResource(this.NS + "state"), aState);
    else { // if (this.getVisitedState(aResource) && aState == "") {
      var property = this.RDF.GetResource(this.NS + "state");
      var stateResource = dataSource.GetTarget(aResource, property, true);

      if (stateResource instanceof Ci.nsIRDFNode)
        dataSource.Unassert(aResource, property, stateResource);
    }
    // if (aSkipParent == undefined)
      // this.hasNewArticles(aResource, true, null);
  },

  getVisitedState: function(aResource)
  {
    var dataSource = this.initDataSource();
    return (dataSource.HasAssertion(aResource, 
                                    this.RDF.GetResource(this.NS + "state"), 
                                    this.RDF.GetLiteral("visited"), 
                                    true));
  },

  getGlobalRefreshState: function(aForcedRefreshFlag)
  {
    if (aForcedRefreshFlag) {
      var dataSource = this.initDataSource();
      var feedSubscriptions = this.getFeedSubscriptions();
      this.refreshStateCache = 0;

      for (i in feedSubscriptions) {
        // dump("\n" + i + " " + feedSubscriptions[i].url);
        if (dataSource.HasAssertion(this.RDF.GetResource(feedSubscriptions[i].urn), 
                                    this.RDF.GetResource(this.NS + "state"), this.RDF.GetLiteral("new"), true)) {
          
          this.refreshStateCache++;
        }
      }
    }
    return this.refreshStateCache;
  },

  getFeedInfoFromDocument: function(aXMLDocument)
  {
    return this.getMostRecentBrowserWindow().mzFeedHandler.getFeedInfoFromDocument(aXMLDocument);
  },

  getDataCollectionFromDocument: function(aXMLDocument, aDataCollection)
  {
    return this.getMostRecentBrowserWindow().mzFeedHandler.getDataCollectionFromDocument(aXMLDocument, aDataCollection);
  },

  getFeedCategoryName: function(aCategoryResource)
  {
    var dataSource = this.initDataSource();
    var data = dataSource.GetTarget(aCategoryResource, 
                                    this.RDF.GetResource(this.NS + "name"), 
                                    true);

    if (data instanceof Ci.nsIRDFLiteral)
      return data.QueryInterface(Ci.nsIRDFLiteral).Value;
    return null;
  },

  renameFeedCategory: function(aCategoryResource, aNewName)
  {
    if (aCategoryResource && /^urn:feedgroup:/.test(aCategoryResource.Value)) {
      var dataSource = this.subscriptionsDS;
      var property = this.RDF.GetResource(this.NS + "name");
      this.addOrChangeProperty(dataSource, aCategoryResource, property, aNewName)
      return aNewName;
    }
    else
      debug("RenameFeedCategory() failed for " + aCategoryResource);
    return null;
  },

  removeFeedCategory: function(aCategoryResource, aPurgeFlag)
  {
    if (aCategoryResource && /^urn:feedgroup:/.test(aCategoryResource.Value)) {
      if (aPurgeFlag == undefined)
        aPurgeFlag = false;

      var dataSource = this.initDataSource();

      if (aCategoryResource instanceof Ci.nsIRDFResource) {
        var categoryContainer = this.RDFCUtils.MakeSeq(dataSource, aCategoryResource);

        if (categoryContainer) {
          var checkbox = { value:aPurgeFlag };
          var checkboxText = null;
          var feedCount = categoryContainer.GetCount();
          var hasFeeds = (feedCount > 0);

          if (hasFeeds)
            checkboxText = this._bundle.GetStringFromName("removeFeedConfirmationCheckboxText");

          var promptService = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
          var navigatorWindow = this.getMostRecentBrowserWindow();
          var categoryName = this.getFeedCategoryName(aCategoryResource);
          var confirmationDialogTitle = this._bundle.GetStringFromName("removeFeedConfirmationTitle");
          var confirmationDialogText = this._bundle.formatStringFromName("removeFeedConfirmationText", [categoryName], 1);
          var confirmResult = promptService.confirmEx(navigatorWindow, confirmationDialogTitle, confirmationDialogText,
                                                      (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_YES) + 
                                                      (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_CANCEL),
                                                      null, null, null, checkboxText, checkbox);

          if (confirmResult == 0) { // Yes returns 0 / No, Esc and X returns 1
            if (hasFeeds) {
              var feeds = this.getResourcesFromContainer(aCategoryResource);
              var currentFeedResource = this.RDF.GetResource(feeds[i]);

              if (checkbox.value) {
                for (i in feeds) {
                  this.moveFeed(currentFeedResource, 
                                aCategoryResource, 
                                this.RDF.GetResource("urn:feeds:root"), -1);
                }
              }
              else {
                for (i in feeds) {
                  this.removeResources(dataSource, currentFeedResource); // remove articles
                  this.removeDataResources(dataSource, currentFeedResource, true); // remove article container
                  this.removeFeedFromFeedCache(currentFeedResource);
                }
              }
            }
            this.removeDataResources(dataSource, aCategoryResource, true); // remove category
            var rootContainer = this.RDFCUtils.MakeSeq(dataSource, this.RDF.GetResource("urn:feeds:root"));
            rootContainer.RemoveElement(aCategoryResource, true);
            this.flushDataSource(dataSource);
            var categoryCacheIndex = this.categoryCache.indexOf(aCategoryResource);

            if (categoryCacheIndex != -1)
              this.categoryCache.splice(categoryCacheIndex, 1);
            return true;
          }
        }
      }
    }
    return false;
  },

  removeResources: function(aDataSource, aResource)
  {
    var container = this.RDFCUtils.MakeSeq(aDataSource, aResource).GetElements();

    while (container.hasMoreElements()) {
      var currentResource = container.getNext().QueryInterface(Ci.nsIRDFResource);
      var labels = aDataSource.ArcLabelsOut(currentResource);

      while (labels.hasMoreElements()) {
        var label = labels.getNext().QueryInterface(Ci.nsIRDFResource);
        var data = aDataSource.GetTarget(currentResource, label, true);

        if (data instanceof Ci.nsIRDFNode)
          aDataSource.Unassert(currentResource, label, data);
      }
    }
    this.flushDataSource(aDataSource);
  },

  removeDataResources: function(aDataSource, aResource, aUnassertLiteralsFlag)
  {
    var resource = this._GetResource(aResource);
    var labels = aDataSource.ArcLabelsOut(resource);

    while (labels.hasMoreElements()) {
      var label = labels.getNext().QueryInterface(Ci.nsIRDFResource);
      var data = aDataSource.GetTarget(resource, label, true);

      if (data instanceof Ci.nsIRDFResource) 
        aDataSource.Unassert(resource, label, data);
      else if (aUnassertLiteralsFlag && data instanceof Ci.nsIRDFLiteral)
        aDataSource.Unassert(resource, label, data);
    }
    this.flushDataSource(aDataSource);
  },

  addFeedToCategory: function(aFeedResource, aCategoryResource)
  {
    this.moveFeed(aFeedResource, 
                  this.RDF.GetResource("urn:feeds:root"), 
                  aCategoryResource, -1);
  },

  removeFeedFromCategory: function(aFeedResource, aCategoryResource)
  {
    this.moveFeed(aFeedResource, 
                  aCategoryResource, 
                  this.RDF.GetResource("urn:feeds:root"), -1);
  },

  isNewFeedURL: function(aURL)
  {
    var dataSource = this.initDataSource();
    var rootResource = this.RDF.GetResource("urn:feeds:root")
    var rootContainer = this.RDFCUtils.MakeSeq(dataSource, rootResource);

    if (!rootContainer) {
      var newContainer = Cc["@mozilla.org/rdf/container;1"].createInstance(Ci.nsIRDFContainer);
      rootContainer = newContainer.Init(dataSource, rootResource);
      return true;
    }
    else {
      aURL = aURL.replace(/www\./,'');
      var uri = this._nsIIOService.newURI(aURL, null, null);
      var urn = "urn:" + uri.host + ":feed-";
      var property = this.RDF.GetResource(this.NS + "url");
      var feedSubscriptions = this.getFeedSubscriptions();

      for (i in feedSubscriptions) {
        if (aURL.indexOf(feedSubscriptions[i].url) == 0) {
          // dump("\nThis feed is already filed!");
          return false;
        }
      }
    }
    // dump("\nThis feed isn't filed yet!");
    return true;
  },

  removeFeed: function(aCategoryResource, aFeedResource)
  {
    var dataSource = this.initDataSource();
    var feedContainer = this.RDFCUtils.MakeSeq(dataSource, aCategoryResource);

    if (feedContainer) {
      feedContainer.Init(dataSource, aCategoryResource);
      feedContainer.RemoveElement(aFeedResource, true);
      this.flushDataSource(dataSource);
      this.removeFeedFromFeedCache(aFeedResource);
      return true;
    }
    return false;
  },

  moveFeed: function(aFeedResource, aFromContainerResource, aToContainerResource, aTargetIndex)
  {
    try {

    var dataSource = this.initDataSource();
    var fromContainer = this.RDFCUtils.MakeSeq(dataSource, aFromContainerResource);
    var toContainer;

    if (aFromContainerResource == aToContainerResource)
      toContainer = fromContainer;
    else 
      toContainer = this.RDFCUtils.MakeSeq(dataSource, aToContainerResource);

    fromContainer.RemoveElement(aFeedResource, true);

    var itemCount = toContainer.GetCount();

    if (aTargetIndex < 1 || aTargetIndex > itemCount)
      toContainer.AppendElement(aFeedResource);
    else
      toContainer.InsertElementAt(aFeedResource, aTargetIndex, true);

    this.flushDataSource(dataSource);
    return aFeedResource;

    } catch(ex) {
      dump("\nmoveFeed - ex: " + ex);
    }
    return null;
  },

  addFeedToFeedCache: function(aFeedResource, aURL)
  {
    if (this.feedCache == null)
      this.feedCache = new Array();

    var feedResourceString = aFeedResource.Value;

    for (i in this.feedCache) {
      if (this.feedCache[i].url == aURL)
        return;
    }
    this.feedCache.push({urn:feedResourceString, url:aURL});
  },

  removeFeedFromFeedCache: function(aFeedResource)
  {
    for (var i = 0; i < this.feedCache.length; i++ ) {
      if (this.feedCache[i].urn == aFeedResource.Value) {
        this.feedCache.splice(i, 1);
        return true;
      }
    }
    return false;
  },

  getFeedInfoFromRDF: function(aInfoArray)
  {
    var names = [ "urn", "feed", "url", "title", "link", "description", "date", "lastModified", 
                  "image", "imageUrl", "imageTitle", "imageWidth", "imageHeight" ];
    var dataSource = this.initDataSource();
    var resource = this._GetResource(aInfoArray[0])
    var feedContainer = this.RDFCUtils.MakeSeq(dataSource, resource);

    if (feedContainer) {
      var lastItem = names.length;

      for (var i = 1; i < lastItem; i++) {
        var data = dataSource.GetTarget(resource, this._GetResource(this.NS + names[i]), true);      
        aInfoArray.push("");

        if (data instanceof Ci.nsIRDFLiteral) {
          var textValue = data.QueryInterface(Ci.nsIRDFLiteral).Value;
          // dump("\nnames[" + i + "]: " + textValue);
          if (i == 1)
            aInfoArray[i] = (textValue.match(/^rss/)) ? "rss" : textValue;
          else if (i == 6 || i == 7)
            aInfoArray[i] = (textValue == "") ? "" : new Date().toString(textValue);
          else
            aInfoArray[i] = textValue;
        }
        if ((data == null) && (i == 8)) {
          // dump("\ndata[" + i + "] : " + data);
          lastItem = names.length;
        }
      }
      return aInfoArray;
    }
    return null;
  },

  getNextURNForHost: function(aHost)
  {
    aHost = aHost.replace(/^urn:/, '').replace(/:feed-(\d*)/, '');
    var indexNumber = 1;
    var currentList = new Array();

    for (var i = 0; i < this.feedCache.length; i++) {
      if (this.feedCache[i].urn.indexOf(aHost) == 4)
        currentList.push(this.feedCache[i].urn.replace(/\D*/, ''));
    }
    if (currentList.length) {
      var gaps = new Array();
      var targetIndex = 1;

      currentList.sort();

      for (var x = 0; x < currentList.length; x++) {
        targetIndex = (x + 1);

        if (currentList[x] != targetIndex)
          gaps.push(targetIndex);
      }
      if (gaps.length)
        indexNumber = gaps[0];
      else {
        indexNumber = (++targetIndex);
      }
    }
    return ("urn:" + aHost + ":feed-" + String(indexNumber));
  },

  exportOPMLData: function(aSkipFilePickerFlag)
  {
    var file;

    if (aSkipFilePickerFlag) {
      // file = 
    }
    else {
      var nsIFilePicker = Ci.nsIFilePicker;
      var filePicker = Cc["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
      var navigatorWindow = this.getMostRecentBrowserWindow();
      filePicker.init(navigatorWindow, this._bundle.GetStringFromName("exportOPML"), nsIFilePicker.modeSave);

      filePicker.appendFilter(this._bundle.GetStringFromName("OPMLFiles"), "*.opml");
      filePicker.appendFilter(this._bundle.GetStringFromName("XMLFiles"), "*.opml; *.xml; *.rdf; *.rss; *.html; *.htm");
      filePicker.appendFilter(this._bundle.GetStringFromName("AllFiles"), "*");

      filePicker.defaultString = "mzWebFeeds.opml";
		
      if (filePicker.show() == nsIFilePicker.returnCancel)
        return false;

      file = filePicker.file;
    }
    var names = [ "xmlUrn", "type", "xmlUrl", "title", "xmlHome", "text", "date", "lastModified", 
                  "image", "imageUrl", "imageTitle", "imageWidth", "imageHeight" ];

    var str = '<?xml version="1.0" encoding="UTF-8"?>\n' +
              '<opml version="2.0">\n' +
              '<head>\n' +
              '\t<title>MultiZilla OPML Export v0.1</title>\n' +
              '\t<dateCreated>' + new Date().toString() + '</dateCreated>\n' +
              '</head>\n' +
              '<body>\n';

    var outputStream = Cc['@mozilla.org/network/file-output-stream;1'].createInstance(Ci.nsIFileOutputStream);
    outputStream.init(file, 0x04 | 0x08 | 0x20, 420, 0 ); 
    outputStream.write(str, str.length);

    var serializer = this._nsIDOMSerializer;
    var feedCategories = this.getFeedCategories();
    var hasFeedCategories = (feedCategories.length > 0);
    var serializedFeeds = new Array();
    var _this = this;

    function _dataCollector(aFeedResource)
    {
      var feedInfo = new Array();
      feedInfo[0] = aFeedResource;
      feedInfo = _this.getFeedInfoFromRDF(feedInfo);

      if (feedInfo) {
        var outline = navigatorWindow.document.createElementNS(null, "outline");

        for (var n = 0; n < feedInfo.length; n++) {
          if (feedInfo[n] != "")
            outline.setAttribute(names[n], feedInfo[n]);
        }
        str = serializer.serializeToString(outline);
        return '\t' + str + '\n';
      }
      return '';
    }

    for (i in feedCategories) {
      var feedCategory = feedCategories[i];
      var feeds = this.getFeedsByCategory(this.RDF.GetResource(feedCategory));

      str = '\t<outline text="' + this.getFeedCategoryName(this.RDF.GetResource(feedCategory)) + '">\n';
      outputStream.write(str, str.length);

      for (i in feeds) {
        var feed = feeds[i];
        serializedFeeds.push(feed);
        str = '\t' + _dataCollector(feed);
        outputStream.write(str, str.length);
      }

      str = '\t</outline>\n';
      outputStream.write(str, str.length);
    }

    for (i in this.feedCache) {
      if (serializedFeeds.indexOf(this.feedCache[i].urn) == -1) {
        str = _dataCollector(this.feedCache[i].urn);
        outputStream.write(str, str.length);
      }
    }
    str = '</body>\n</opml>';
	outputStream.write(str, str.length);
    outputStream.flush();
    outputStream.close();
    return true;
  },

  importOPMLData: function(aSkipFilePickerFlag)
  {
    debug("Not Yet Implemented!");
  },

  rebuildStorage: function()
  {
    //this.exportOPMLData(true);
    this.importOPMLData(true);
    this.storageNeedsRebuild = false;
  },

  storeFeed: function(aFeedObject)
  {
    if (this._getPref("offline", false)) {
      var file = this.getProfile();
      var hostName = aFeedObject.urn.replace(/(urn:)?(:feed-\w*)?/g, "");
      var subdirs = ["extensions", "multizilla@multizilla.mozdev.org", "feeds", hostName];

      for (i in subdirs) {
        file.append(subdirs[i]);

        if (!file.exists() || !file.isDirectory())
          file.create(Ci.nsIFile.DIRECTORY_TYPE, 0664);
      }
      var fileName = aFeedObject.urn.replace(/(urn:)?(\w*\.)?(\w*:)?/g, "");
      var xmldoc = aFeedObject.document;
      // var rss = xmldoc.getElementsByTagName("rss");
      fileExtension = "xml"; // (rss instanceof Ci.nsIDOMHTMLCollection) ? "rss": "xml";
      file.append(fileName + "." + fileExtension);
      var stream = this._nsInsIFileOutputStream;
      var serializer = this._nsIDOMSerializer;
      stream.init(file, 0x02 | 0x08 | 0x20, 0664, 0);
      serializer.serializeToStream(xmldoc, stream, "");
      stream.close();
    }
  },

  getFeedJarURI: function(aFeedURN)
  {
    var profileURI = this.getProfileURI();
    var profileURL = profileURI.spec + "extensions/multizilla@multizilla.mozdev.org/feeds/";

    if (aFeedURN) {
      profileURL += aFeedURN.replace(/(urn:)?(:feed-\w*)?/g, "");
      profileURL += "/" + aFeedURN.replace(/(urn:)?(\w*\.)?(\w*:)?/g, "") + ".xml";
    }
    return this._nsIIOService.newURI(profileURL, null, null);
  },

  getProfile: function()
  {
    var file = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties);
    return file.get("ProfD", Ci.nsILocalFile);
  },

  getProfileURI: function()
  {
    var profileDir = this.getProfile();
    return this._nsIIOService.newFileURI(profileDir);
  },

  getVersionInfo: function()
  {
    return VERSION_INFO;
  },

  getFeedCategories: function()
  {
    if (this.categoryCache == null) { // is the category cache already initialized?
      var dataSource = this.initDataSource();
      var rootContainer = this.RDFCUtils.MakeSeq(dataSource, this.RDF.GetResource("urn:feeds:root")).GetElements();

      if (!rootContainer)
        return null;

      var feedGroups = new Array();

      while (rootContainer.hasMoreElements()) {
        var resource = rootContainer.getNext().QueryInterface(Ci.nsIRDFResource).Value;

        if (/^urn:feedgroup:/.test(resource))
          feedGroups.push(resource);
      }
      this.categoryCache = feedGroups; // initialize the category cache
      return (feedGroups.length) ? feedGroups : null;
    }
    return this.categoryCache; // return cached entries
  },

  getFeedsByCategory: function(aContainerResource)
  {
    return this.getResourcesFromContainer(aContainerResource);
  },

  getCategoryByFeed: function(aFeedResource)
  {
    var feedCategories = this.getFeedCategories();
    var dataSource = this.subscriptionsDS; // initialized by 'getFeedCategoryies'

    if (feedCategories) {
      for (i in feedCategories) {
        var resourceString = feedCategories[i];
        var resource = this.RDF.GetResource(resourceString);
        var container = this.RDFCUtils.MakeSeq(dataSource, resource);

        if (container.IndexOf(aFeedResource) >= 0)
          return resourceString;
      }
    }
    var urn = "urn:feeds:root";
    var rootContainer = this.RDFCUtils.MakeSeq(dataSource, this.RDF.GetResource(urn));

    if (rootContainer && rootContainer.IndexOf(aFeedResource) >= 0)
      return urn;

    return null;
  },

  getFeedSubscriptions: function()
  {
    if (this.feedCache == null) { // is the feed cache already initialized?
      this.feedCache = new Array();
      var dataSource = this.initDataSource();
      var feedList = this.getResourcesFromContainer(this.RDF.GetResource("urn:feeds:root"));
      var property = this.RDF.GetResource(this.NS + "url");

      for (i in feedList) {
        var url, urn = feedList[i];
        var feedResource = this.RDF.GetResource(urn);
        var urlNode = dataSource.GetTarget(feedResource, property, true);

        if (urlNode instanceof Ci.nsIRDFNode) {
          url = urlNode.QueryInterface(Ci.nsIRDFLiteral).Value.replace(/www\./,'');

          if (url)
            this.feedCache.push({urn:feedList[i], url: url});
        }
        else {
          // Remove empty/junk feeds (the "Subscribe to feed" button relies on this)
          var category = this.getCategoryByFeed(feedResource);
          var categoryResource = this.RDF.GetResource(category);
          // dump("\nRetrieving of: " + urn + " in category: " + category + " failed. Purging...");
          this.removeFeed(categoryResource, feedResource);
          this.flushDataSource(dataSource);
        }
      }
      if (this.storageNeedsRebuild)
        this.rebuildStorage();
    }
    return this.feedCache;
  },

  getResourcesFromContainer: function(aContainerResource)
  {
    var _this = this;
    var resourceList = new Array();
    var duplicatedFeedCache = new Array();
    var dataSource = this.initDataSource();

    function _inner(aContainerResource) {
      var category = aContainerResource.Value;
      var container = _this.RDFCUtils.MakeSeq(dataSource, aContainerResource);

      if (!container)
        return -1;

      var categoryContainer = container.GetElements();

      while (categoryContainer.hasMoreElements()) {
        try {
          var resource = categoryContainer.getNext().QueryInterface(Ci.nsIRDFResource);
          var urn = resource.Value;

          if (/^urn:feedgroup:/.test(urn))
            _inner(resource);
          else {
            if (resourceList.indexOf(urn) == -1)
              resourceList.push(urn);
            else {
              dump("\nDuplicated feed: \"" + urn + "\" found in category: " + category + " Purging...");
              _this.storageNeedsRebuild = true;
              duplicatedFeedCache.unshift({container:container, resource:resource});
            }
          }
        } catch(ex) {
          dump("\ngetResourcesFromContainer - ex: " + ex);
        }
      }
      return (resourceList.length) ? resourceList : null; // .sort();
    }
    resourceList = _inner(aContainerResource);

    for (index in duplicatedFeedCache) {
      var targetFeed = duplicatedFeedCache[index];
      targetFeed.container.RemoveElement(targetFeed.resource, true);
    }
    this.flushDataSource(dataSource);

    return resourceList;
  },

  getMostRecentBrowserWindow: function()
  {
    return this._getMostRecentWindow("navigator:browser");
  },

  getOfflineStatus: function()
  {
    return Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService).offline;
  },

  addOrChangeProperty: function(aDataSource, aResource, aProperty, aValue)
  {
    var currentValue = aDataSource.GetTarget(aResource, aProperty, true) || null;
    var newValue = this.RDF.GetLiteral(aValue);

    this.addOrChangeValue(aDataSource, aResource, aProperty, currentValue, newValue);
  },

  addOrChangeValue: function(aDataSource, aResource, aProperty, aCurrentValue, aValue)
  {
    if (aCurrentValue)
      aDataSource.Change(aResource, aProperty, aCurrentValue, aValue);
    else
      aDataSource.Assert(aResource, aProperty, aValue, true);
    this.flushDataSource(aDataSource);
  },

  flushDataSource: function(aDataSource)
  {
    try {
      aDataSource.QueryInterface(Ci.nsIRDFRemoteDataSource).Flush();
    } catch(ex) {
      // die silently
    }
  },

  _GetResource: function(aResource)
  {
    if (typeof aResource == "string")
      aResource = this.RDF.GetResource(aResource);
    return aResource;
  },

  _getMostRecentWindow: function(aWindowType)
  {
    return this._nsIWindowMediator.getMostRecentWindow(aWindowType);
  },

  _getPref: function(aName, aDefault) {
    var prefBranch = this.prefBranch;

    try {
      switch (prefBranch.getPrefType(aName))
      {
        case prefBranch.PREF_STRING:
             return prefBranch.getCharPref(aName);
        case prefBranch.PREF_BOOL:
             return prefBranch.getBoolPref(aName);
        case prefBranch.PREF_INT:
             return prefBranch.getIntPref(aName);
        default:
             return aDefault;
      }
    }
    catch(ex) {
      return aDefault;
    }
  },

  QueryInterface: function(aIID)
  {
    if (aIID.equals(Ci.nsIWebFeedService) ||
        aIID.equals(Ci.nsIObserver) ||
        aIID.equals(Ci.nsIProfileStartupListener) ||
        aIID.equals(Ci.nsISupports))
      return this;

    Components.returnCode = Cr.NS_ERROR_NO_INTERFACE;
    return null;
  }
};

/* :::::::: Service Registration & Initialization ::::::::::::::: */

/* ........ nsIModule .............. */

const mzWebFeedModule = 
{
  mClassName: "MultiZilla Web Feed Service",
  mContractID: "@multizilla.org/webfeed-service;1",
  mClassID: Components.ID("3d4f753c-eb6c-11db-8314-0800200c9a66"),

  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aCID.equals(this.mClassID))
      throw Cr.NS_ERROR_NO_INTERFACE;
    if (!aIID.equals(Ci.nsIFactory))
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;

    return this.mFactory;
  },

  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    debug("Registering...\n");
    aCompMgr = aCompMgr.QueryInterface(Ci.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(this.mClassID, this.mClassName, this.mContractID, aFileSpec, aLocation, aType);
    // this.getCategoryManager().addCategoryEntry("profile-startup-category", this.mContractID, "", true, true);

    // Make the Web Feed Service a startup observer
    this.getCategoryManager().addCategoryEntry("app-startup", this.mClassName,
                                               "service," + this.mContractID, 
                                               true, true, null);
  },

  unregisterSelf: function(aCompMgr, aFileSpec, aLocation)
  {
    debug("Unregistering...\n");
    aCompMgr = aCompMgr.QueryInterface(Ci.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(this.mClassID, aFileSpec);
    this.getCategoryManager().deleteCategoryEntry("app-startup", this.mContractID, true);
  },

  canUnload: function(aCompMgr)
  {
    return true;
  },

  getCategoryManager: function()
  {
    return Cc["@mozilla.org/categorymanager;1"].getService(Ci.nsICategoryManager);
  },

  /* ........ nsIFactory .............. */

  mFactory:
  {
    createInstance: function(aOuter, aIID)
    {
      if (aOuter != null)
        throw Cr.NS_ERROR_NO_AGGREGATION;
      if (!aIID.equals(Ci.nsIWebFeedService) && 
          !aIID.equals(Ci.nsIObserver) &&
          !aIID.equals(Ci.nsIProfileStartupListener) &&
          !aIID.equals(Ci.nsISupports))
        throw Cr.NS_ERROR_INVALID_ARG;

      return mzWebFeedService.QueryInterface(aIID);
    },

    lockFactory: function(aLock)
    {
      // quiten warnings
    }
  }
};

function NSGetModule(aCompMgr, aFileSpec)
{
  return mzWebFeedModule;
}

